from decimal import Decimal
from io import BytesIO


from chellow.e.bill_parsers.haven_edi import (
    BillElement,
    Parser,
    _process_BTL,
    _process_CCD1,
    _process_CCD3,
    _process_CLO,
    _process_MAN,
    _process_MHD,
    _process_MTR,
)
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeasurementRequirement,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcLlfcSsc,
    MtcLlfcSscPc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Pc,
    Site,
    Source,
    Ssc,
    Tpr,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_process_BTL_zeroes(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "03", "nhh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "",
        False,
        False,
        meter_type,
        meter_payment_type,
        2,
        vf,
        None,
    )
    ssc = Ssc.insert(sess, "0154", "All", True, vf, None)
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, vf, None)
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, vf, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2000, 1, 31, 23, 30),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        "0154",
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    elements = {
        "UVLT": ["0"],
        "UTVA": ["0"],
        "TBTL": ["0"],
    }
    headers = {
        "errors": [],
        "issue_date": to_utc(ct_datetime(2024, 1, 1)),
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "mpan_core": "22 7867 6232 781",
        "reads": [],
        "breakdown": {},
        "bill_elements": [
            BillElement(
                gbp=Decimal("10.31"),
                rate=Decimal("0.0001"),
                cons=Decimal("113"),
                titles=None,
                desc="Night",
            )
        ],
        "sess": sess,
        "reference": "ref 01",
        "kwh": Decimal("0"),
        "account": "acc",
        "bill_type_code": "N",
    }
    bill = _process_BTL(elements, headers)
    expected_headers = {
        "errors": [],
        "issue_date": to_utc(ct_datetime(2024, 1, 1)),
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "mpan_core": "22 7867 6232 781",
        "breakdown": {},
        "reads": [],
        "sess": sess,
        "bill_elements": [
            BillElement(
                gbp=Decimal("10.31"),
                rate=Decimal("0.0001"),
                cons=Decimal("113"),
                titles=None,
                desc="Night",
            ),
        ],
        "reference": "ref 01",
        "kwh": Decimal("0"),
        "account": "acc",
        "bill_type_code": "N",
    }
    assert headers == expected_headers
    assert str(bill["net"]) == "0.00"
    expected_bill = {
        "breakdown": {},
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "gross": Decimal("0.00"),
        "issue_date": to_utc(ct_datetime(2024, 1, 1)),
        "mpan_core": "22 7867 6232 781",
        "net": Decimal("0.00"),
        "reads": [],
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "vat": Decimal("0.00"),
        "elements": [
            {
                "breakdown": {
                    "kwh": Decimal("113"),
                    "rate": {Decimal("0.0001")},
                },
                "start_date": to_utc(ct_datetime(2025, 1, 1)),
                "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
                "name": "00221",
                "net": Decimal("10.31"),
            },
        ],
        "reference": "ref 01",
        "kwh": Decimal("0"),
        "account": "acc",
        "bill_type_code": "N",
    }
    assert bill == expected_bill


def test_process_BTL_non_zeroes(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "03", "nhh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "",
        False,
        False,
        meter_type,
        meter_payment_type,
        2,
        vf,
        None,
    )
    ssc = Ssc.insert(sess, "0154", "All", True, vf, None)
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, vf, None)
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, vf, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2000, 1, 31, 23, 30),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        "0154",
        energisation_status,
        dtc_meter_type,
        "22 7368 2392 188",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    elements = {
        "UVLT": ["11"],
        "UTVA": ["12"],
        "TBTL": ["10"],
    }
    headers = {
        "errors": [],
        "issue_date": to_utc(ct_datetime(2024, 1, 1)),
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "elements": [],
        "mpan_core": "22 7368 2392 188",
        "reads": [],
        "breakdown": {},
        "sess": sess,
        "bill_elements": [
            BillElement(
                gbp=Decimal("10.31"),
                rate=Decimal("0.0001"),
                cons=Decimal("113"),
                titles=None,
                desc="Night",
            ),
        ],
        "sess": sess,
        "reference": "ref 01",
        "kwh": Decimal("0"),
        "bill_type_code": "N",
        "account": "acc",
    }
    bill = _process_BTL(elements, headers)
    expected_headers = {
        "errors": [],
        "issue_date": to_utc(ct_datetime(2024, 1, 1)),
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "elements": [],
        "mpan_core": "22 7368 2392 188",
        "reads": [],
        "breakdown": {},
        "bill_elements": [
            BillElement(
                gbp=Decimal("10.31"),
                rate=Decimal("0.0001"),
                cons=Decimal("113"),
                titles=None,
                desc="Night",
            ),
        ],
        "sess": sess,
        "reference": "ref 01",
        "kwh": Decimal("0"),
        "bill_type_code": "N",
        "account": "acc",
    }
    expected_bill = {
        "kwh": Decimal("0"),
        "net": Decimal("0.11"),
        "vat": Decimal("0.12"),
        "gross": Decimal("0.10"),
        "breakdown": {},
        "elements": [
            {
                "breakdown": {
                    "kwh": Decimal("113"),
                    "rate": {Decimal("0.0001")},
                },
                "name": "00221",
                "net": Decimal("10.31"),
                "start_date": to_utc(ct_datetime(2025, 1, 1)),
                "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
            }
        ],
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "issue_date": to_utc(ct_datetime(2024, 1, 1)),
        "mpan_core": "22 7368 2392 188",
        "reads": [],
        "reference": "ref 01",
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "bill_type_code": "N",
        "account": "acc",
    }
    assert headers == expected_headers
    assert bill == expected_bill


def test_process_MTR_UTLHDR(mocker):
    elements = {}
    headers = {"message_type": "UTLHDR", "breakdown": {}}
    bill = _process_MTR(elements, headers)
    assert bill is None


def test_process_MTR_UTLBIL(mocker):
    elements = {}
    headers = {}
    expected_headers = {}
    _process_MTR(elements, headers)
    assert headers == expected_headers


def test_process_BTL_multiple_charges_one_tpr(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "03", "nhh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "",
        False,
        False,
        meter_type,
        meter_payment_type,
        2,
        vf,
        None,
    )
    ssc = Ssc.insert(sess, "0244", "All", True, vf, None)
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, vf, None)
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, vf, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2000, 1, 31, 23, 30),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        "0244",
        energisation_status,
        dtc_meter_type,
        "22 7368 2392 188",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    gbp = "10.31"
    cons = "113"
    elements = {
        "UVLT": ["11"],
        "UTVA": ["12"],
        "TBTL": ["10"],
    }
    headers = {
        "sess": sess,
        "message_type": "UTLBIL",
        "breakdown": {},
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
        ],
        "mpan_core": "22 7368 2392 188",
        "kwh": 8,
        "reference": "a",
        "issue_date": to_utc(ct_datetime(2025, 1, 1)),
        "account": "acc",
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "net": 0,
        "vat": 0,
        "gross": 0,
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "Day",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            },
        ],
        "bill_type_code": "N",
        "errors": [],
    }
    bill = _process_BTL(elements, headers)
    expected_headers = {
        "mpan_core": "22 7368 2392 188",
        "errors": [],
        "sess": sess,
        "message_type": "UTLBIL",
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Night",
            ),
        ],
        "kwh": 8,
        "reference": "a",
        "issue_date": to_utc(ct_datetime(2025, 1, 1)),
        "account": "acc",
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "net": 0,
        "vat": 0,
        "gross": 0,
        "breakdown": {},
        "reads": [
            {
                "msn": "hgkh",
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "00040",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ],
        "bill_type_code": "N",
    }
    assert headers == expected_headers
    expected_bill = {
        "breakdown": {},
        "elements": [
            {
                "breakdown": {
                    "kwh": Decimal("113"),
                    "rate": {Decimal("0.0001")},
                },
                "start_date": to_utc(ct_datetime(2025, 1, 1)),
                "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
                "name": "00206",
                "net": Decimal("10.31"),
            },
            {
                "breakdown": {
                    "kwh": Decimal("113"),
                    "rate": {Decimal("0.0001")},
                },
                "start_date": to_utc(ct_datetime(2025, 1, 1)),
                "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
                "name": "00206",
                "net": Decimal("10.31"),
            },
        ],
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "gross": Decimal("0.10"),
        "issue_date": to_utc(ct_datetime(2025, 1, 1)),
        "mpan_core": "22 7368 2392 188",
        "net": Decimal("0.11"),
        "reads": [
            {
                "coefficient": Decimal("0.00001"),
                "mpan": "      ",
                "msn": "hgkh",
                "pres_date": to_utc(ct_datetime(2020, 3, 1, 23, 30)),
                "pres_type_code": "N",
                "pres_value": Decimal("0"),
                "prev_date": to_utc(ct_datetime(2020, 3, 31, 23, 30)),
                "prev_type_code": "N",
                "prev_value": Decimal("1"),
                "tpr_code": "00040",
                "units": "kWh",
            },
        ],
        "vat": Decimal("0.12"),
        "kwh": 8,
        "reference": "a",
        "account": "acc",
        "bill_type_code": "N",
    }
    assert bill == expected_bill


def test_process_BTL_unmetered(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "",
        False,
        False,
        meter_type,
        meter_payment_type,
        2,
        vf,
        None,
    )
    tpr = Tpr.insert(sess, "00001", True, True)
    ssc = Ssc.insert(sess, "0393", "All", True, vf, None)
    MeasurementRequirement.insert(sess, ssc, tpr)
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, vf, None)
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, vf, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2000, 1, 31, 23, 30),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        None,
        "22 7368 2392 188",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    gbp = "10.31"
    cons = "113"
    elements = {
        "UVLT": ["11"],
        "UTVA": ["12"],
        "TBTL": ["10"],
    }
    headers = {
        "errors": [],
        "sess": sess,
        "breakdown": {},
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Energy Charges",
            )
        ],
        "mpan_core": "22 7368 2392 188",
        "kwh": 8,
        "reference": "a",
        "issue_date": to_utc(ct_datetime(2025, 1, 1)),
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "account": "acc",
        "reads": [],
        "bill_type_code": "N",
    }
    bill = _process_BTL(elements, headers)
    expected_headers = {
        "errors": [],
        "sess": sess,
        "bill_elements": [
            BillElement(
                gbp=Decimal(gbp),
                rate=Decimal("0.0001"),
                cons=Decimal(cons),
                titles=None,
                desc="Energy Charges",
            )
        ],
        "kwh": 8,
        "reference": "a",
        "mpan_core": "22 7368 2392 188",
        "account": "acc",
        "issue_date": to_utc(ct_datetime(2025, 1, 1)),
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "breakdown": {},
        "reads": [],
        "bill_type_code": "N",
    }
    assert headers == expected_headers
    expected_bill = {
        "breakdown": {},
        "kwh": 8,
        "reference": "a",
        "mpan_core": "22 7368 2392 188",
        "account": "acc",
        "issue_date": to_utc(ct_datetime(2025, 1, 1)),
        "start_date": to_utc(ct_datetime(2025, 1, 1)),
        "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
        "net": Decimal("0.11"),
        "vat": Decimal("0.12"),
        "gross": Decimal("0.10"),
        "elements": [
            {
                "name": "00001",
                "net": Decimal("10.31"),
                "breakdown": {
                    "rate": {Decimal("0.0001")},
                    "kwh": Decimal("113"),
                },
                "start_date": to_utc(ct_datetime(2025, 1, 1)),
                "finish_date": to_utc(ct_datetime(2025, 1, 31, 23, 30)),
            }
        ],
        "reads": [],
        "bill_type_code": "N",
    }
    assert bill == expected_bill


def test_process_MAN(mocker):
    elements = {
        "MADN": ["20", "0000000000", "6", "00", "001", "002"],
    }

    headers = {}
    _process_MAN(elements, headers)
    expected_headers = {"mpan_core": "20 0000 0000 006"}
    assert headers == expected_headers


def test_process_MHD(mocker):
    message_type = "UTLBIL"
    elements = {"TYPE": [message_type]}

    sess = mocker.Mock()
    headers = {"sess": sess}
    _process_MHD(elements, headers)
    expected_headers = {
        "message_type": message_type,
        "reads": [],
        "errors": [],
        "bill_elements": [],
        "breakdown": {"raw-lines": []},
        "sess": sess,
        "kwh": Decimal("0"),
    }
    assert headers == expected_headers
    assert type(headers["breakdown"]) is type(expected_headers)


def test_process_CCD3(mocker):
    elements = {
        "CCDE": ["3", "", "NRG"],
        "TCOD": ["NIGHT", "Night"],
        "TMOD": ["453043"],
        "CONS": [[]],
        "BPRI": ["10"],
    }

    headers = {
        "bill_elements": [],
        "kwh": Decimal("0"),
    }

    _process_CCD3(elements, headers)

    expected_headers = {
        "kwh": Decimal("0"),
        "bill_elements": [
            BillElement(
                gbp=Decimal("0.00"),
                rate=Decimal("0.0001"),
                cons=Decimal("0"),
                titles=None,
                desc="Night",
            ),
        ],
    }
    assert headers == expected_headers


def test_process_CCD_1(mocker):
    msn = "hgkh"

    elements = {
        "CCDE": ["1", "", "NRG"],
        "TCOD": ["NIGHT", "Night"],
        "TMOD": ["453043"],
        "MTNR": [msn],
        "CONS": [[]],
        "BPRI": ["10"],
        "PRDT": ["200301"],
        "PVDT": ["200331"],
        "MLOC": [""],
        "PRRD": ["0", "00", "1", "00"],
        "ADJF": ["", "1"],
    }

    headers = {
        "reads": [],
    }

    _process_CCD1(elements, headers)

    expected_headers = {
        "reads": [
            {
                "msn": msn,
                "mpan": "      ",
                "coefficient": Decimal("0.00001"),
                "units": "kWh",
                "tpr_code": "453043",
                "prev_date": utc_datetime(2020, 3, 31, 22, 30),
                "prev_value": Decimal("1"),
                "prev_type_code": "N",
                "pres_date": utc_datetime(2020, 3, 1, 23, 30),
                "pres_value": Decimal("0"),
                "pres_type_code": "N",
            }
        ]
    }
    assert headers == expected_headers


def test_process_CLO(mocker):
    account = "accnt"

    elements = {"CLOC": [account, ""]}

    headers = {}

    _process_CLO(elements, headers)

    expected_headers = {
        "account": "",
    }
    assert headers == expected_headers


def test_make_raw_bills(sess):
    file_lines = [
        "MHD=2+UTLBIL:3'",
        "MTR=6'",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))

    parser = Parser(f)
    parser.make_raw_bills()


def test_Parser(mocker, sess):
    file_lines = [
        "STX=ANA:1+3832776:Fusion PLC+771:Fusion PLC+130214:090426+8472+"
        "PASSWORD+UTLHDR'",
        "END=251'",
    ]
    f = BytesIO(b"\n".join(n.encode("utf8") for n in file_lines))
    parser = Parser(f)
    parser.make_raw_bills()



from decimal import Decimal

import xlrd.sheet

from chellow.e.bill_parsers.sww_xls import _parse_row
from chellow.utils import ct_datetime, to_utc


def test_parse_row(mocker):
    row = []
    for val in [
        "2019-04-01",
        "2019-04-30",
        "22 7673 2894 214",
        54,
        7.8,
        74,
        5998,
        6.22,
        52,
        877,
        8.001,
        82.98,
        377,
        0.3,
        52,
        823,
        0.782,
        305,
        672,
        4.62,
        8483,
        6843,
        6.328,
        6984,
        7034,
        6734,
        574,
        956,
        678,
        4567,
        485,
        57,
        76,
        39647,
        6587,
        768,
        34687,
        67,
        5478,
        879654,
        654897,
        54789,
        None,
        None,
        None,
        43895,
        48,
        498,
        3498,
        9384,
        23,
    ]:
        cell = mocker.Mock(spec=xlrd.sheet.Cell)
        cell.value = val
        row.append(cell)

    row_index = 2
    datemode = 0
    titles = [
        "BILL START DATE",
        "BILL END DATE",
        "MPAN",
        "DUOS-AVAIL-AV USE KVADAY",
        "DUOS-AVAIL-AV RATE P/KVADAY",
        "DUOS-AVAIL-AV COST GBP",
        "DUOS-STND-SITE USE DAY",
        "DUOS-STND-SITE RATE P/DAY",
        "DUOS-STND-SITE COST GBP",
        "DUOS-UNIT-AMBER USE KWH",
        "DUOS-UNIT-AMBER RATE P/KWH",
        "DUOS-UNIT-AMBER COST GBP",
        "DUOS-UNIT-GREEN USE KWH",
        "DUOS-UNIT-GREEN RATE P/KWH",
        "DUOS-UNIT-GREEN COST GBP",
        "DUOS-UNIT-RED USE KWH",
        "DUOS-UNIT-RED RATE P/KWH",
        "DUOS-UNIT-RED COST GBP",
        "LVY-AAHEDC-ALL USE KWH",
        "LVY-AAHEDC-ALL RATE P/KWH",
        "LVY-AAHEDC-ALL COST GBP",
        "LVY-BSUOS-ALL USE KWH",
        "LVY-BSUOS-ALL RATE P/KWH",
        "LVY-BSUOS-ALL COST GBP",
        "LVY-CCL-ALL USE KWH",
        "LVY-CCL-ALL RATE P/KWH",
        "LVY-CCL-ALL COST GBP",
        "LVY-CFDOB-ALL USE KWH",
        "LVY-CFDOB-ALL RATE P/KWH",
        "LVY-CFDOB-ALL COST GBP",
        "LVY-CFDOP-ALL USE KWH",
        "LVY-CFDOP-ALL RATE P/KWH",
        "LVY-CFDOP-ALL COST GBP",
        "LVY-FIT-ALL USE KWH",
        "LVY-FIT-ALL RATE P/KWH",
        "LVY-FIT-ALL COST GBP",
        "LVY-RO-ALL USE KWH",
        "LVY-RO-ALL RATE P/KWH",
        "LVY-RO-ALL COST GBP",
        "NRG-UNIT-SUMMER USE KWH",
        "NRG-UNIT-SUMMER RATE P/KWH",
        "NRG-UNIT-SUMMER COST GBP",
        "NRG-UNIT-WINTER USE KWH",
        "NRG-UNIT-WINTER RATE P/KWH",
        "NRG-UNIT-WINTER COST GBP",
        "SYS-ADMIN-ALL USE MO",
        "SYS-ADMIN-ALL RATE GBP/MO",
        "SYS-ADMIN-ALL COST GBP",
        "TUOS-TRIAD-SITE USE DAY",
        "TUOS-TRIAD-SITE RATE GBP/DAY",
        "TUOS-TRIAD-SITE COST GBP",
    ]
    title_row = []
    for title in titles:
        cell = mocker.Mock(spec=xlrd.sheet.Cell)
        cell.value = title
        title_row.append(cell)

    issue_date = to_utc(ct_datetime(2019, 1, 1))

    bill = _parse_row(issue_date, row, row_index, datemode, title_row)
    assert bill == {
        "account": "2007",
        "bill_type_code": "N",
        "breakdown": {},
        "elements": [
            {
                "name": "duos-availability",
                "breakdown": {
                    "kvaday": Decimal("54"),
                    "rate": Decimal("7.8"),
                },
                "net": Decimal("74.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "duos-fixed",
                "breakdown": {
                    "day": Decimal("5998"),
                    "rate": Decimal("6.22"),
                },
                "net": Decimal("52.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "duos-amber",
                "breakdown": {
                    "kwh": Decimal("877"),
                    "rate": Decimal("8.001"),
                },
                "net": Decimal("82.98"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "duos-green",
                "breakdown": {
                    "kwh": Decimal("377"),
                    "rate": Decimal("0.3"),
                },
                "net": Decimal("52.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "duos-red",
                "breakdown": {
                    "kwh": Decimal("823"),
                    "rate": Decimal("0.782"),
                },
                "net": Decimal("305.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "aahedc",
                "breakdown": {
                    "kwh": Decimal("672"),
                    "rate": Decimal("4.62"),
                },
                "net": Decimal("8483.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "bsuos",
                "breakdown": {
                    "kwh": Decimal("6843"),
                    "rate": Decimal("6.328"),
                },
                "net": Decimal("6984.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "ccl",
                "breakdown": {
                    "kwh": Decimal("7034"),
                    "rate": Decimal("6734"),
                },
                "net": Decimal("574.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "cfdob",
                "breakdown": {
                    "kwh": Decimal("956"),
                    "rate": Decimal("678"),
                },
                "net": Decimal("4567.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "cfdop",
                "breakdown": {
                    "kwh": Decimal("485"),
                    "rate": Decimal("57"),
                },
                "net": Decimal("76.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "fit",
                "breakdown": {
                    "kwh": Decimal("39647"),
                    "rate": Decimal("6587"),
                },
                "net": Decimal("768.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "ro",
                "breakdown": {
                    "kwh": Decimal("34687"),
                    "rate": Decimal("67"),
                },
                "net": Decimal("5478.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "summer",
                "breakdown": {
                    "kwh": Decimal("879654"),
                    "rate": Decimal("654897"),
                },
                "net": Decimal("54789.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "admin",
                "breakdown": {
                    "mo": Decimal("43895"),
                    "rate": Decimal("48"),
                },
                "net": Decimal("498.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
            {
                "name": "triad",
                "breakdown": {
                    "day": Decimal("3498"),
                    "rate": Decimal("9384"),
                },
                "net": Decimal("23.00"),
                "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
                "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
            },
        ],
        "finish_date": to_utc(ct_datetime(2019, 4, 30, 23, 30)),
        "gross": Decimal("82805.98"),
        "issue_date": issue_date,
        "kwh": Decimal("672"),
        "mpan_core": "22 0003 0354 632",
        "net": Decimal("82805.98"),
        "reads": [],
        "reference": "20190101T0000_3",
        "start_date": to_utc(ct_datetime(2019, 4, 1, 0, 0)),
        "vat": Decimal("0.00"),
    }



from collections import defaultdict
from decimal import Decimal

from chellow.e.computer import SiteSource, SupplySource
from chellow.e.duos import datum_2010_04_01, datum_2012_02_23, duos_vb
from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcLlfcSsc,
    MtcLlfcSscPc,
    MtcParticipant,
    MtcSsc,
    Participant,
    Pc,
    Site,
    Source,
    Ssc,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, hh_range, to_utc, utc_datetime


def test_duos_vb_unmetered(sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    caches = {"dno": {"22": {}}}
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)

    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    Ssc.insert(sess, "0393", "All", True, vf, None)
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2000, 1, 31, 23, 30),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        None,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    dno_rates = {
        "_L": {
            "bands": [],
            "ums_bands": [
                {
                    "band": "black",
                    "weekend": False,
                    "start-decimal-hour": Decimal(0.5),
                    "finish-decimal-hour": Decimal(1),
                    "start-month": 1,
                    "finish-month": 12,
                },
                {
                    "band": "yellow",
                    "weekend": False,
                    "start-decimal-hour": Decimal(1),
                    "finish-decimal-hour": Decimal(1.5),
                    "start-month": 1,
                    "finish-month": 12,
                },
            ],
            "tariffs": {
                "510": {
                    "description": "Unmetered Supplies",
                    "gbp-per-kvarh": 0,
                    "black-gbp-per-kwh": Decimal("0.83"),
                    "green-gbp-per-kwh": Decimal("0.72"),
                    "yellow-gbp-per-kwh": Decimal("0.32"),
                    "gbp-per-mpan-per-day": 0,
                    "gbp-per-kva-per-day": 0,
                }
            },
        }
    }
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, vf, None, dno_rates)

    start_date = to_utc(ct_datetime(2019, 2, 1))
    finish_date = to_utc(ct_datetime(2019, 2, 1, 1, 0))

    for hh in hh_range(caches, start_date, finish_date):
        llfc.insert_laf(sess, hh, 1)

    era = supply.eras[0]

    ds = SupplySource(
        sess,
        start_date,
        finish_date,
        to_utc(ct_datetime(2025, 1, 1)),
        era,
        True,
        caches,
    )

    sess.commit()

    duos_vb(ds)

    assert ds.hh_data == [
        {
            "anti-msp-kw": 0,
            "anti-msp-kwh": 0,
            "ct-day": 1,
            "ct-day-of-week": 4,
            "ct-decimal-hour": 0.0,
            "ct-is-bank-holiday": False,
            "ct-is-month-end": False,
            "ct-month": 2,
            "ct-year": 2019,
            "duos-description": "Unmetered Supplies",
            "duos-reactive-kvarh": 0.0,
            "duos-green-gbp": 0.0,
            "duos-green-kwh": 0,
            "duos-green-rate": 0.72,
            "exp-msp-kvar": 0,
            "exp-msp-kvarh": 0,
            "gsp-kw": 0.0,
            "gsp-kwh": 0.0,
            "hist-export-grid-kvarh": 0,
            "hist-imp-msp-kvarh": 0,
            "hist-import-grid-kvarh": 0,
            "hist-kwh": 0,
            "hist-start": utc_datetime(2019, 2, 1, 0, 0),
            "imp-msp-kvar": 0,
            "imp-msp-kvarh": 0,
            "laf": 1.0,
            "msp-kva": 0.0,
            "msp-kw": 0,
            "msp-kwh": 0,
            "start-date": utc_datetime(2019, 2, 1, 0, 0),
            "status": "X",
            "utc-day": 1,
            "utc-day-of-week": 4,
            "utc-decimal-hour": 0.0,
            "utc-hour": 0,
            "utc-is-bank-holiday": False,
            "utc-is-month-end": False,
            "utc-minute": 0,
            "utc-month": 2,
            "utc-year": 2019,
        },
        {
            "anti-msp-kw": 0,
            "anti-msp-kwh": 0,
            "ct-day": 1,
            "ct-day-of-week": 4,
            "ct-decimal-hour": 0.5,
            "ct-is-bank-holiday": False,
            "ct-is-month-end": False,
            "ct-month": 2,
            "ct-year": 2019,
            "duos-description": "Unmetered Supplies",
            "duos-reactive-kvarh": 0.0,
            "duos-black-gbp": 0.0,
            "duos-black-kwh": 0,
            "duos-black-rate": 0.83,
            "exp-msp-kvar": 0,
            "exp-msp-kvarh": 0,
            "gsp-kw": 0.0,
            "gsp-kwh": 0.0,
            "hist-export-grid-kvarh": 0,
            "hist-imp-msp-kvarh": 0,
            "hist-import-grid-kvarh": 0,
            "hist-kwh": 0,
            "hist-start": utc_datetime(2019, 2, 1, 0, 30),
            "imp-msp-kvar": 0,
            "imp-msp-kvarh": 0,
            "laf": 1.0,
            "msp-kva": 0.0,
            "msp-kw": 0,
            "msp-kwh": 0,
            "start-date": utc_datetime(2019, 2, 1, 0, 30),
            "status": "X",
            "utc-day": 1,
            "utc-day-of-week": 4,
            "utc-decimal-hour": 0.5,
            "utc-hour": 0,
            "utc-is-bank-holiday": False,
            "utc-is-month-end": False,
            "utc-minute": 30,
            "utc-month": 2,
            "utc-year": 2019,
        },
        {
            "anti-msp-kw": 0,
            "anti-msp-kwh": 0,
            "ct-day": 1,
            "ct-day-of-week": 4,
            "ct-decimal-hour": 1.0,
            "ct-is-bank-holiday": False,
            "ct-is-month-end": False,
            "ct-month": 2,
            "ct-year": 2019,
            "duos-description": "Unmetered Supplies",
            "duos-reactive-kvarh": 0.0,
            "duos-yellow-gbp": 0.0,
            "duos-yellow-kwh": 0,
            "duos-yellow-rate": 0.32,
            "exp-msp-kvar": 0,
            "exp-msp-kvarh": 0,
            "gsp-kw": 0.0,
            "gsp-kwh": 0.0,
            "hist-export-grid-kvarh": 0,
            "hist-imp-msp-kvarh": 0,
            "hist-import-grid-kvarh": 0,
            "hist-kwh": 0,
            "hist-start": utc_datetime(2019, 2, 1, 1, 0),
            "imp-msp-kvar": 0,
            "imp-msp-kvarh": 0,
            "laf": 1.0,
            "msp-kva": 0.0,
            "msp-kw": 0,
            "msp-kwh": 0,
            "start-date": utc_datetime(2019, 2, 1, 1, 0),
            "status": "X",
            "utc-day": 1,
            "utc-day-of-week": 4,
            "utc-decimal-hour": 1.0,
            "utc-hour": 1,
            "utc-is-bank-holiday": False,
            "utc-is-month-end": False,
            "utc-minute": 0,
            "utc-month": 2,
            "utc-year": 2019,
        },
    ]


def test_duos_availability_from_to(mocker, sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    caches = {"dno": {"22": {}}}
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", to_utc(ct_datetime(2000, 1, 1)), None, "22"
    )
    dno_contract = Contract.insert_dno(
        sess,
        dno.dno_code,
        participant,
        "",
        {},
        valid_from,
        None,
        {},
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(1996, 1, 1)),
        None,
    )

    month_from = to_utc(ct_datetime(2019, 2, 1))
    month_to = to_utc(ct_datetime(2019, 2, 28, 23, 30))

    for hh in hh_range(caches, month_from, month_to):
        llfc.insert_laf(sess, hh, 1)

    sess.commit()

    ds = mocker.Mock()
    ds.dno_code = "22"
    ds.dno_contract = dno_contract
    ds.gsp_group_code = "_L"
    ds.llfc_code = "510"
    ds.is_displaced = False
    ds.sc = 0
    ds.supplier_bill = defaultdict(int)
    ds.supplier_rate_sets = defaultdict(set)
    ds.get_data_sources = mocker.Mock(return_value=[])
    ds.caches = caches
    ds.sess = sess
    dno_rates = {
        "_L": {
            "bands": {},
            "tariffs": {
                "510": {
                    "gbp-per-kvarh": 0,
                    "green-gbp-per-kwh": 0,
                    "gbp-per-mpan-per-day": 0,
                    "gbp-per-kva-per-day": 0,
                }
            },
        }
    }
    ds.hh_rate = mocker.Mock(return_value=dno_rates)

    hh = {
        "start-date": ct_datetime(2019, 2, 28, 23, 30),
        "ct-decimal-hour": 23.5,
        "ct-is-month-end": True,
        "ct-day-of-week": 3,
        "ct-year": 2019,
        "ct-month": 2,
        "msp-kwh": 0,
        "imp-msp-kvarh": 0,
        "exp-msp-kvarh": 0,
    }
    datum_2010_04_01(ds, hh)

    ds.get_data_sources.assert_called_with(month_from, month_to)


def test_lafs_hist(mocker, sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    caches = {"dno": {"22": {}}}
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    insert_voltage_levels(sess)
    vl = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(sess, "510", "5", vl, False, True, valid_from, None)

    hist_laf = 1.5
    start_date = to_utc(ct_datetime(2019, 2, 28, 23, 30))
    hist_date = to_utc(ct_datetime(2018, 2, 28, 23, 30))
    llfc.insert_laf(sess, hist_date, hist_laf)

    sess.commit()

    ds = mocker.Mock()
    ds.start_date = start_date
    ds.finish_date = start_date
    ds.history_start = hist_date
    ds.history_finish = hist_date
    ds.dno_code = "22"
    ds.gsp_group_code = "_L"
    ds.llfc_code = "510"
    ds.is_displaced = False
    ds.sc = 0
    ds.supplier_bill = defaultdict(int)
    ds.supplier_rate_sets = defaultdict(set)
    ds.get_data_sources = mocker.Mock(return_value=iter([ds]))
    ds.caches = caches
    ds.sess = sess
    ds.non_primary_elements = set()
    dno_rates = {
        "_L": {
            "bands": {},
            "tariffs": {
                "510": {
                    "description": "LV Sub Generation Site Specific",
                    "gbp-per-kvarh": 0,
                    "green-gbp-per-kwh": 0,
                    "gbp-per-mpan-per-day": 0,
                    "gbp-per-kva-per-day": 0,
                }
            },
        }
    }
    ds.hh_rate = mocker.Mock(return_value=dno_rates)

    hh = {
        "hist-start": hist_date,
        "start-date": start_date,
        "ct-decimal-hour": 23.5,
        "ct-is-month-end": True,
        "ct-day-of-week": 3,
        "ct-year": 2019,
        "ct-month": 2,
        "msp-kwh": 0,
        "msp-kw": 0,
        "msp-kva": 0,
        "imp-msp-kvarh": 0,
        "imp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "exp-msp-kvar": 0,
    }
    ds.hh_data = [hh]
    datum_2012_02_23(ds, hh)

    assert caches["dno"]["22"]["lafs"]["510"][start_date] == hist_laf


def test_lafs_forecast_none(mocker, sess):
    dno_code = "22"
    caches = {"dno": {dno_code: {}}}
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(
        sess, market_role_R, "WPD", to_utc(ct_datetime(2000, 1, 1)), None, "22"
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc_code = "510"
    dno.insert_llfc(
        sess,
        llfc_code,
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        to_utc(ct_datetime(1996, 1, 1)),
        None,
    )

    start_date = to_utc(ct_datetime(2019, 2, 28, 23, 30))
    hist_date = to_utc(ct_datetime(2018, 2, 28, 23, 30))

    sess.commit()

    ds = mocker.Mock()
    ds.forecast_date = hist_date
    ds.start_date = start_date
    ds.finish_date = start_date
    ds.history_start = hist_date
    ds.history_finish = hist_date
    ds.dno_code = dno_code
    ds.gsp_group_code = "_L"
    ds.llfc_code = llfc_code
    ds.is_displaced = False
    ds.sc = 0
    ds.supplier_bill = defaultdict(int)
    ds.supplier_rate_sets = defaultdict(set)
    ds.get_data_sources = mocker.Mock(return_value=iter([ds]))
    ds.caches = caches
    ds.sess = sess
    ds.non_primary_elements = set()
    dno_rates = {
        "_L": {
            "bands": {},
            "tariffs": {
                llfc_code: {
                    "description": "LV Sub Generation Site Specific",
                    "gbp-per-kvarh": 0,
                    "green-gbp-per-kwh": 0,
                    "gbp-per-mpan-per-day": 0,
                    "gbp-per-kva-per-day": 0,
                }
            },
        }
    }
    ds.hh_rate = mocker.Mock(return_value=dno_rates)

    hh = {
        "hist-start": hist_date,
        "start-date": start_date,
        "ct-decimal-hour": 23.5,
        "ct-is-month-end": True,
        "ct-day-of-week": 3,
        "ct-year": 2019,
        "ct-month": 2,
        "msp-kwh": 0,
        "msp-kw": 0,
        "msp-kva": 0,
        "imp-msp-kvarh": 0,
        "imp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "exp-msp-kvar": 0,
    }
    ds.hh_data = [hh]
    datum_2012_02_23(ds, hh)

    assert caches["dno"][dno_code]["lafs"][llfc_code][start_date] == 1


def test_lafs_forecast(mocker, sess):
    valid_from = to_utc(ct_datetime(2000, 1, 1))
    caches = {"dno": {"22": {}}}
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    insert_voltage_levels(sess)
    vl = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(sess, "510", "HV", vl, False, True, valid_from, None)

    forecast_date = to_utc(ct_datetime(2018, 5, 31, 23, 30))
    llfc.insert_laf(sess, forecast_date, 1.4)

    sess.commit()

    hist_date = to_utc(ct_datetime(2018, 2, 28, 23, 30))
    start_date = to_utc(ct_datetime(2019, 2, 28, 23, 30))

    ds = mocker.Mock()
    ds.start_date = start_date
    ds.finish_date = start_date
    ds.history_start = hist_date
    ds.history_finish = hist_date
    ds.forecast_date = forecast_date
    ds.dno_code = "22"
    ds.gsp_group_code = "_L"
    ds.llfc_code = "510"
    ds.is_displaced = False
    ds.sc = 0
    ds.supplier_bill = defaultdict(int)
    ds.supplier_rate_sets = defaultdict(set)
    ds.get_data_sources = mocker.Mock(return_value=iter([ds]))
    ds.caches = caches
    ds.sess = sess
    ds.non_primary_elements = set()
    dno_rates = {
        "_L": {
            "bands": {},
            "tariffs": {
                "510": {
                    "description": "LV Sub Generation Site Specific",
                    "gbp-per-kvarh": 0,
                    "green-gbp-per-kwh": 0,
                    "gbp-per-mpan-per-day": 0,
                    "gbp-per-kva-per-day": 0,
                }
            },
        }
    }
    ds.hh_rate = mocker.Mock(return_value=dno_rates)

    hh = {
        "hist-start": hist_date,
        "start-date": start_date,
        "ct-decimal-hour": 23.5,
        "ct-is-month-end": True,
        "ct-day-of-week": 3,
        "ct-year": 2019,
        "ct-month": 2,
        "msp-kwh": 0,
        "msp-kw": 0,
        "msp-kva": 0,
        "imp-msp-kvarh": 0,
        "imp-msp-kvar": 0,
        "exp-msp-kvarh": 0,
        "exp-msp-kvar": 0,
    }
    ds.hh_data = [hh]
    datum_2012_02_23(ds, hh)


def test_SiteSource(sess):
    valid_from = utc_datetime(1996, 1, 1)
    site = Site.insert(sess, "CI017", "Water Works")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(
        sess, market_role_Z, "None core", utc_datetime(2000, 1, 1), None, None
    )
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(
        sess, market_role_M, "Fusion Mop Ltd", valid_from, None, None
    )
    participant.insert_party(sess, market_role_X, "Fusion Ltc", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, valid_from, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, valid_from, None, {}
    )
    pc = Pc.insert(sess, "03", "nhh", utc_datetime(2000, 1, 1), None)
    ssc = Ssc.insert(sess, "0393", "unrestricted", True, utc_datetime(2000, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    dno_rates = {
        "tariffs": {
            "510": {
                "day-gbp-per-kwh": 0,
                "night-gbp-per-kwh": 0,
                "reactive-gbp-per-kvarh": 0,
            }
        },
        "lafs": {"hv": {"other": 0}},
    }
    Contract.insert_dno(
        sess, dno.dno_code, participant, "", {}, valid_from, None, dno_rates
    )
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, True, True, valid_from, None
    )
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "D")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    mtc_ssc = MtcSsc.insert(sess, mtc_participant, ssc, valid_from, None)
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    mtc_llfc_ssc = MtcLlfcSsc.insert(sess, mtc_ssc, llfc, valid_from, None)
    MtcLlfcSscPc.insert(sess, mtc_llfc_ssc, pc, valid_from, None)
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        ssc.code,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    active_channel = era.insert_channel(sess, False, "ACTIVE")
    active_data_raw = [
        {
            "start_date": utc_datetime(2009, 8, 10),
            "value": 10,
            "status": "A",
        }
    ]
    active_channel.add_hh_data(sess, active_data_raw)
    era.insert_channel(sess, True, "REACTIVE_IMP")

    sess.commit()

    caches = {}
    start_date = utc_datetime(2009, 7, 31, 23, 00)
    finish_date = utc_datetime(2009, 8, 31, 22, 30)
    forecast_date = utc_datetime(2019, 8, 31, 22, 30)
    ss = SiteSource(sess, site, start_date, finish_date, forecast_date, caches, era=era)
    duos_vb(ss)



import csv
import threading
import traceback

from flask import g, redirect


from sqlalchemy import null, or_, select
from sqlalchemy.orm import joinedload

from chellow.dloads import open_file
from chellow.models import Batch, Bill, Contract, Era, RSession, ReportRun, User
from chellow.utils import (
    c_months_u,
    csv_make_val,
    hh_max,
    hh_min,
    hh_range,
    req_date,
    req_int,
    to_ct,
)


def content(user_id, report_run_id, contract_id, months_length, finish_date):
    f = writer = None
    try:
        with RSession() as sess:
            caches = {}
            contract = Contract.get_by_id(sess, contract_id)
            user = User.get_by_id(sess, user_id)
            f = open_file(
                f"missing_bills_{contract.id}.csv", user, mode="w", newline=""
            )
            writer = csv.writer(f, lineterminator="\n")
            titles = (
                "contract_name",
                "month_start",
                "month_finish",
                "site_code",
                "site_name",
                "imp_mpan_core",
                "exp_mpan_core",
                "account",
            )
            writer.writerow(titles)

            finish_date_ct = to_ct(finish_date)

            for month_start, month_finish in c_months_u(
                finish_year=finish_date_ct.year,
                finish_month=finish_date_ct.month,
                months=months_length,
            ):
                missing_bills = {}
                missing_account = {}
                account_missing_tuple = hh_range(caches, month_start, month_finish)
                for era in sess.scalars(
                    select(Era)
                    .where(
                        Era.start_date <= month_finish,
                        or_(Era.finish_date == null(), Era.finish_date >= month_start),
                        or_(
                            Era.mop_contract == contract,
                            Era.dc_contract == contract,
                            Era.imp_supplier_contract == contract,
                            Era.exp_supplier_contract == contract,
                        ),
                    )
                    .options(joinedload(Era.supply))
                ):
                    chunk_start = hh_max(era.start_date, month_start)
                    chunk_finish = hh_min(era.finish_date, month_finish)
                    missing_set = set(hh_range(caches, chunk_start, chunk_finish))
                    if era.mop_contract == contract:
                        account = ""
                    elif era.dc_contract == contract:
                        account = ""
                    elif era.imp_supplier_contract == contract:
                        account = era.imp_supplier_account
                    elif era.imp_supplier_contract == contract:
                        account = era.exp_supplier_account

                    try:
                        account_missing_set = missing_account[account]
                    except KeyError:
                        account_missing_set = missing_account[account] = set(
                            account_missing_tuple
                        )
                    supply = era.supply

                    for bill in sess.scalars(
                        select(Bill)
                        .join(Batch)
                        .where(
                            Batch.contract == contract,
                            Bill.supply == supply,
                            Bill.start_date <= chunk_finish,
                            Bill.finish_date >= chunk_start,
                        )
                    ):
                        found_set = set(
                            hh_range(caches, bill.start_date, bill.finish_date)
                        )
                        missing_set.difference_update(found_set)
                        account_missing_set.difference_update(found_set)

                    if len(missing_set) > 0:
                        site = era.get_physical_site(sess)

                        values = {
                            "contract_id": contract.id,
                            "contract_name": contract.name,
                            "month_start": month_start,
                            "month_finish": month_finish,
                            "era_id": era.id,
                            "supply_id": supply.id,
                            "imp_mpan_core": era.imp_mpan_core,
                            "exp_mpan_core": era.exp_mpan_core,
                            "site_id": site.id,
                            "site_code": site.code,
                            "site_name": site.name,
                            "account": account,
                            "market_role_code": contract.market_role.code,
                        }
                        missing_bills[era.id] = values
                for era_id, values in missing_bills.items():
                    if len(missing_account[values["account"]]) > 0:
                        writer.writerow(csv_make_val(values[t]) for t in titles)
                        ReportRun.w_insert_row(report_run_id, "", titles, values, {})

    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        if writer is not None:
            writer.writerow([msg])
    finally:
        ReportRun.w_update(report_run_id, "finished")
        if f is not None:
            f.close()


def do_get(sess):
    contract_id = req_int("contract_id")
    months = req_int("months")
    finish_date = req_date("finish", resolution="month")
    report_run = ReportRun.insert(
        sess,
        "missing_e_bills",
        g.user,
        "missing_e_bills",
        {},
    )
    sess.commit()
    args = g.user.id, report_run.id, contract_id, months, finish_date
    threading.Thread(target=content, args=args).start()
    return redirect(f"/report_runs/{report_run.id}", 303)



from collections import namedtuple
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.edi_lib import parse_edi, to_date, to_gbp
from chellow.models import Session, Ssc, Supply
from chellow.utils import HH, parse_mpan_core, to_ct, to_utc

READ_TYPE_MAP = {
    "00": "N",
    "01": "Q",
    "02": "E",
    "03": "F",
    "04": "C",
    "05": "M",
    "06": "I",
}

SSC_MAP = {
    "0008": {
        "Other": "00159",
    },
    "0065": {
        "Other": "00010",
    },
    "0036": {
        "Other": "00151",
    },
    "0037": {
        "Other": "00153",
    },
    "0151": {
        "Day": "00043",
        "Night": "00210",
    },
    "0154": {
        "Day": "00039",
        "Night": "00221",
    },
    "0174": {"Day": "01071", "Night": "01072"},
    "0179": {"Day": "01139", "Night": "01140"},
    "0184": {"Day": "01149", "Night": "01150"},
    "0186": {"Day": "01153", "Night": "01154"},
    "0242": {"Day": "00044", "Night": "00208"},
    "0244": {"Day": "00040", "Night": "00206"},
    "0246": {
        "Night": "00160",
        "Other": "00277",
        "Weekday": "00276",
    },
    # 0246
    "22 0002 1401 756": {
        "Night": "00277",
        "Other": "00160",
        "Weekday": "00276",
    },
    "0252": {
        "Other": "00212",
    },
    "0265": {
        "Other": "00190",
    },
    "0319": {
        "Other": "00071",
        "Weekday": "00183",
    },
    "0320": {
        "Other": "00072",
        "Weekday": "00184",
    },
    "0322": {
        "Weekday": "01073",
        "Other": "01074",
    },
    "0326": {
        "Night": "00210",
        "Other": "00187",
        "Weekday": "00184",
        "Day": "00184",
    },
    "0393": {
        "Day": "00001",
        "Energy Charges": "00001",
    },
    "0428": {
        "Energy Charges": "00258",
        "Energy Charges 2": "00259",
        "Day": "00259",
    },
}


# None denotes a TPR-based charge

TMOD_MAP = {
    "700285": ("standing", "rate", "days"),
    "422733": ("ccl", "rate", "kwh"),
    "066540": ("ccl", "rate", "kwh"),
    "453043": None,
    "493988": ("reconciliation", None, None),
    "068476": None,
    "265091": None,
    "517180": None,
    "547856": None,
}


BillElement = namedtuple("BillElement", ["gbp", "rate", "cons", "titles", "desc"])


def _process_BCD(elements, headers):
    headers["issue_date"] = to_date(elements["IVDT"][0])
    headers["reference"] = elements["INVN"][0]
    headers["bill_type_code"] = elements["BTCD"][0]

    sumo = elements["SUMO"]
    headers["start_date"] = to_date(sumo[0])
    headers["finish_date"] = to_date(sumo[1]) + relativedelta(days=1) - HH


def _to_date(component):
    return to_utc(to_ct(Datetime.strptime(component, "%y%m%d")))


def _to_decimal(components, divisor=None):
    comp_0 = components[0]
    if comp_0 == "":
        return None

    try:
        result = Decimal(comp_0)
    except InvalidOperation as e:
        raise BadRequest(f"Can't parse '{comp_0}' of {components} as a decimal: {e}")

    if len(components) > 1 and components[-1] == "R":
        result *= Decimal("-1")

    if divisor is not None:
        result /= Decimal(divisor)

    return result


def _process_NOOP(elements, headers):
    pass


def _process_BTL(elements, headers):
    sess = headers["sess"]
    try:
        mpan_core = headers["mpan_core"]
    except KeyError:
        raise BadRequest("The mpan_core can't be found for this bill.")

    start_date = headers["start_date"]
    reads = headers["reads"]
    supply = Supply.get_by_mpan_core(sess, mpan_core)
    era = supply.find_era_at(sess, start_date)
    if era is None:
        era = supply.find_last_era(sess)

    if era is None:
        imp_mpan_core = ""
        ssc = Ssc.get_by_code(sess, "0393")
    else:
        imp_mpan_core = era.imp_mpan_core
        ssc = Ssc.get_by_code(sess, "0393", start_date) if era.ssc is None else era.ssc

    try:
        ssc_lookup = imp_mpan_core
        tpr_map = SSC_MAP[ssc_lookup]
    except KeyError:
        ssc_lookup = ssc.code
        try:
            tpr_map = SSC_MAP[ssc_lookup]
        except KeyError:
            raise BadRequest(f"The SSC {ssc_lookup} isn't in the SSC_MAP.")

    for read in reads:
        desc = read["tpr_code"]
        try:
            read["tpr_code"] = tpr_map[desc]
        except KeyError:
            raise BadRequest(
                f"The description {desc} isn't in the SSC_MAP for the SSC {ssc_lookup}."
            )

    elems = []
    for el in headers["bill_elements"]:
        if el.titles is None:
            try:
                tpr = tpr_map[el.desc]
            except KeyError:
                raise BadRequest(
                    f"The billing element description {el.desc} isn't in the "
                    f"SSC_MAP for the SSC {ssc_lookup}."
                )

            elname, elrate, elcons = f"{tpr}", "rate", "kwh"
        else:
            elname, elrate, elcons = el.titles

        bd = {}
        if elrate is not None and el.rate is not None:
            bd[elrate] = {el.rate}
        if elcons is not None and el.cons is not None:
            bd[elcons] = el.cons

        if el.gbp is not None:
            elems.append(
                {
                    "name": elname,
                    "net": el.gbp,
                    "breakdown": bd,
                    "start_date": headers["start_date"],
                    "finish_date": headers["finish_date"],
                }
            )

    bill = {
        "net": to_gbp(elements["UVLT"]),
        "vat": to_gbp(elements["UTVA"]),
        "gross": to_gbp(elements["TBTL"]),
        "elements": elems,
        "mpan_core": headers["mpan_core"],
        "reads": headers["reads"],
        "start_date": headers["start_date"],
        "finish_date": headers["finish_date"],
        "issue_date": headers["issue_date"],
        "breakdown": headers["breakdown"],
        "reference": headers["reference"],
        "kwh": headers["kwh"],
        "bill_type_code": headers["bill_type_code"],
        "account": headers["account"],
    }

    if len(headers["errors"]) > 0:
        bill["error"] = " ".join(headers["errors"])
    return bill


def _process_CLO(elements, headers):
    cloc = elements["CLOC"]
    headers["account"] = cloc[1]
    # headers['msn'] = cloc[2] if len(cloc) > 2 else ''


def _process_MAN(elements, headers):
    madn = elements["MADN"]
    dno = madn[0]
    unique = madn[1]
    check_digit = madn[2]
    # pc = madn[3]
    # mtc = madn[4]
    # llfc = madn[5]

    headers["mpan_core"] = parse_mpan_core("".join([dno, unique, check_digit]))


def _process_MTR(elements, headers):
    pass


def _process_MHD(elements, headers):
    message_type = elements["TYPE"][0]
    sess = headers["sess"]
    if message_type == "UTLBIL":
        headers.clear()
        headers["kwh"] = Decimal("0")
        headers["reads"] = []
        headers["breakdown"] = {"raw-lines": []}
        headers["bill_elements"] = []
        headers["errors"] = []
        headers["sess"] = sess
    headers["message_type"] = message_type


def _process_CCD4(elements, headers):
    tmod_1 = elements["TMOD"][0]
    try:
        eln_gbp, eln_rate, eln_cons = TMOD_MAP[tmod_1]
    except KeyError:
        raise BadRequest(
            f"Can't find the Tariff Modifer Code 1 {tmod_1} in the TMOD_MAP."
        )

    """
    m = elements['MLOC'][0]
    mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))
    """
    breakdown = headers["breakdown"]

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = _to_decimal(cons, "1000")
        breakdown[eln_cons] = el_cons

    if eln_rate is not None:
        rate = _to_decimal(elements["BPRI"], "100000")
        try:
            rates = breakdown[eln_rate]
        except KeyError:
            rates = breakdown[eln_rate] = set()

        rates.append(rate)

    """
    start_date = _to_date(elements['CSDT'][0])
    finish_date = _to_date(elements['CEDT'][0]) - HH
    """

    if "CTOT" in elements:
        net = Decimal("0.00") + _to_decimal(elements["CTOT"], "100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net


def _process_CCD1(elements, headers):
    pres_read_date = _to_date(elements["PRDT"][0]) + relativedelta(days=1) - HH

    prev_read_date = _to_date(elements["PVDT"][0]) + relativedelta(days=1) - HH

    m = elements["MLOC"][0]
    mpan = " ".join((m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13]))

    prrd = elements["PRRD"]
    try:
        pres_read_type = READ_TYPE_MAP[prrd[1]]
    except KeyError as e:
        raise BadRequest(f"The present register read type isn't recognized {e}")

    try:
        prev_read_type = READ_TYPE_MAP[prrd[3]]
    except KeyError as e:
        raise BadRequest(f"The previous register read type isn't recognized {e}")

    coefficient = Decimal(elements["ADJF"][1]) / Decimal(100000)
    pres_reading_value = Decimal(prrd[0])
    prev_reading_value = Decimal(prrd[2])
    msn = elements["MTNR"][0]
    tpr_code = elements["TMOD"][0]
    if tpr_code == "kW":
        units = "kW"
        tpr_code = None
    elif tpr_code == "kVA":
        units = "kVA"
        tpr_code = None
    else:
        if tpr_code == "":
            tpr_code = elements["TCOD"][1]

        units = "kWh"

    headers["reads"].append(
        {
            "msn": msn,
            "mpan": mpan,
            "coefficient": coefficient,
            "units": units,
            "tpr_code": tpr_code,
            "prev_date": prev_read_date,
            "prev_value": prev_reading_value,
            "prev_type_code": prev_read_type,
            "pres_date": pres_read_date,
            "pres_value": pres_reading_value,
            "pres_type_code": pres_read_type,
        }
    )


def _process_CCD2(elements, headers):
    tmod_1 = elements["TMOD"][0]
    try:
        titles = TMOD_MAP[tmod_1]
    except KeyError:
        raise BadRequest(
            f"Can't find the Tariff Modifier Code 1 {tmod_1} in the TMOD_MAP."
        )

    """
    m = elements['MLOC'][0]
    mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))
    """

    if tmod_1 == "700285":  # standing charge
        start_date = _to_date(elements["CSDT"][0])
        finish_date = _to_date(elements["CEDT"][0])
        elcons = Decimal((finish_date - start_date).days + 1)
    else:
        cons = elements["CONS"]
        if len(cons[0]) > 0:
            elcons = _to_decimal(cons, "1000")

    rate = _to_decimal(elements["BPRI"], "100000")

    if "CTOT" in elements:
        gbp = Decimal("0.00") + _to_decimal(elements["CTOT"], "100")
    else:
        gbp = Decimal("0.00")

    headers["bill_elements"].append(
        BillElement(gbp=gbp, titles=titles, rate=rate, cons=elcons, desc=None)
    )


def _process_CCD3(elements, headers):
    tcod = elements["TCOD"]

    tmod_1 = elements["TMOD"][0]
    try:
        titles = TMOD_MAP[tmod_1]
    except KeyError:
        raise BadRequest(
            f"Can't find the Tariff Code Modifier 1 {tmod_1} in the TMOD_MAP."
        )

    if len(tcod) == 2:
        desc = tcod[1]
    else:
        desc = None

    """
    m = elements['MLOC'][0]
    mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))
    """
    if tmod_1 == "700285":  # standing charge
        start_date = _to_date(elements["CSDT"][0])
        finish_date = _to_date(elements["CEDT"][0])
        consumption = Decimal((finish_date - start_date).days + 1)
    else:
        cons = elements["CONS"]
        if len(cons[0]) > 0:
            consumption = _to_decimal(cons, "1000")
        else:
            consumption = Decimal("0")

    if titles is None:
        headers["kwh"] += consumption

    rate = _to_decimal(elements["BPRI"], "100000")

    gbp = Decimal("0.00")
    if "CTOT" in elements:
        gbp += _to_decimal(elements["CTOT"], "100")

    headers["bill_elements"].append(
        BillElement(gbp=gbp, rate=rate, cons=consumption, titles=titles, desc=desc)
    )


def _process_VAT(elements, header):
    pass


CODE_FUNCS = {
    "BCD": _process_BCD,
    "BTL": _process_BTL,
    "CCD1": _process_CCD1,
    "CCD2": _process_CCD2,
    "CCD3": _process_CCD3,
    "CCD4": _process_CCD4,
    "CDA": _process_NOOP,
    "CDT": _process_NOOP,
    "CLO": _process_CLO,
    "DNA": _process_NOOP,
    "END": _process_NOOP,
    "FIL": _process_NOOP,
    "MAN": _process_MAN,
    "MHD": _process_MHD,
    "MTR": _process_MTR,
    "REF": _process_NOOP,
    "SDT": _process_NOOP,
    "STX": _process_NOOP,
    "TYP": _process_NOOP,
    "TTL": _process_NOOP,
    "VAT": _process_VAT,
    "VTS": _process_NOOP,
}


def _process_segment(headers, line_number, line, seg_name, elements):
    try:
        func = CODE_FUNCS[seg_name]
    except KeyError:
        raise BadRequest(f"Code {seg_name} not recognized.")

    try:
        bill = func(elements, headers)
    except BadRequest as e:
        raise BadRequest(
            f"{e.description} on line {line_number} line {line} "
            f"seg_name {seg_name} elements {elements}"
        )
    except BaseException as e:
        raise BadRequest(
            f"Problem on line {line_number} line {line} "
            f"seg_name {seg_name} elements {elements}"
        ) from e

    if "breakdown" in headers:
        headers["breakdown"]["raw-lines"].append(line)

    return bill


class Parser:
    def __init__(self, f):
        self.edi_str = str(f.read(), "utf-8", errors="ignore")
        self.line_number = None

    def make_raw_bills(self):
        bills = []
        bill = None
        with Session() as sess:
            headers = {"sess": sess}
            for self.line_number, line, seg_name, elements in parse_edi(self.edi_str):
                bill = _process_segment(
                    headers, self.line_number, line, seg_name, elements
                )
                if bill is not None:
                    bills.append(bill)

        return bills



from sqlalchemy import select

from chellow.models import Supply, User


def make_issue_bundles(sess, issues):
    return [make_issue_bundle(sess, issue) for issue in issues]


def make_issue_bundle(sess, issue):
    props = issue.properties
    supply_bundles = []
    if "owner_id" in props:
        owner = User.get_by_id(sess, props["owner_id"])
    else:
        owner = None
    bundle = {
        "issue": issue,
        "latest_entry": None if len(issue.entries) == 0 else issue.entries[0],
        "supplies": supply_bundles,
        "owner": owner,
    }
    for supply in sess.scalars(
        select(Supply)
        .where(Supply.id.in_(props.get("supply_ids", [])))
        .order_by(Supply.id)
    ).all():
        era = supply.eras[-1]
        site = era.get_physical_site(sess)
        supply_bundles.append({"supply": supply, "era": era, "site": site})
    return bundle



import csv
from codecs import iterdecode
from datetime import datetime as Datetime
from decimal import Decimal

from werkzeug.exceptions import BadRequest

from chellow.utils import (
    HH,
    parse_mpan_core,
    to_ct,
    to_utc,
)

# From the EDF Portal select:
# - Electricity
# - Half-hourly
# - Reactive Power and kWh combined
# export data as CSV
#
#
# "Month","MPAN","Measurement Type","Settlement Date","00:00",..,"00:00_Act_Est",
# "Apr-24", "2200000000000", "AI", "01-04-2024", "0.0000", ..., "Actual", "Actual",


def create_parser(reader, mpan_map, messages):
    return HhParserEdfCsv(reader, mpan_map, messages)


CHANNEL_LOOKUP = {
    "AI": "ACTIVE",
    "RI": "REACTIVE_IMP",
    "RE": "REACTIVE_EXP",
}

STATUS_LOOKUP = {
    "Actual": "A",
    "Estimated": "E",
    "-": "E",
}


def _get_field(values, index, name):
    if len(values) > index:
        return values[index].strip()
    else:
        raise BadRequest(f"Can't find field {index}, {name}.")


def _find_data(shredder):
    try:
        for line_number, values in enumerate(shredder, start=2):
            mpan_core_str = _get_field(values, 1, "MPAN Core")
            mpan_core = parse_mpan_core(mpan_core_str)
            channel_type_str = _get_field(values, 2, "Channel Type")
            channel_type = CHANNEL_LOOKUP[channel_type_str]
            start_day_ct_str = _get_field(values, 3, "Day")
            start_day_ct = to_ct(Datetime.strptime(start_day_ct_str, "%d-%m-%Y"))

            for i in range(50):
                val_idx = i + 4
                value_str = values[val_idx]
                if value_str == "-":
                    break

                status_idx = val_idx + 50
                status_str = values[status_idx]
                try:
                    status = STATUS_LOOKUP[status_str]
                except KeyError:
                    raise BadRequest(
                        f"Unrecognized status {status_str} at index {status_idx}."
                    )
                yield {
                    "mpan_core": mpan_core,
                    "channel_type": channel_type,
                    "start_date": to_utc(start_day_ct + HH * i),
                    "value": Decimal(value_str),
                    "status": status,
                }
    except BadRequest as e:
        e.description = (
            f"Problem at line number: {line_number} : {values} : {e.description}"
        )
        raise e


class HhParserEdfCsv:
    def __init__(self, reader, mpan_map, messages):
        s = iterdecode(reader, "utf-8")
        self.reader = csv.reader(s)
        next(self.reader)  # skip the title line
        self.data = iter(_find_data(self.reader))

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.data)

    def close(self):
        self.shredder.close()



from collections import defaultdict
from decimal import Decimal

from chellow.gas.bill_parser_total_edi import (
    _process_CCD2,
    _process_MHD,
    _to_finish_date,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_to_finish_date():
    actual = _to_finish_date("211031")
    assert actual == to_utc(ct_datetime(2021, 10, 31, 23, 30))


def test_process_MHD():
    message_type = "UTLBIL"
    headers = {}
    elements = {"TYPE": [message_type, "3"]}
    _process_MHD(elements, headers)

    expected_headers = {
        "message_type": message_type,
        "reads": [],
        "raw_lines": [],
        "breakdown": {"units_consumed": Decimal("0")},
        "kwh": Decimal("0.00"),
        "net": Decimal("0.00"),
        "vat": Decimal("0.00"),
        "gross": Decimal("0.00"),
    }
    assert headers == expected_headers


def test_process_CCD2():
    headers = {}
    headers["reads"] = []
    headers["raw_lines"] = []
    headers["breakdown"] = defaultdict(int, {"units_consumed": Decimal(0)})
    headers["kwh"] = Decimal("0.00")
    headers["net"] = Decimal("0.00")
    headers["vat"] = Decimal("0.00")
    headers["gross"] = Decimal("0.00")
    elements = {
        "CCDE": ["", "", "MET"],
        "NUCT": ["968823", "KWH"],
        "MTNR": [],
        "CONB": [],
        "ADJF": [],
        "PRDT": [],
        "PVDT": [],
        "PRRD": [],
        "MLOC": ["978672"],
        "CSDT": ["240401"],
        "CEDT": ["240430"],
        "CTOT": ["3882"],
        "CPPU": ["9742"],
    }
    _process_CCD2(elements, headers)
    print(headers)
    expected_headers = {
        "reads": [],
        "raw_lines": [],
        "breakdown": {
            "units_consumed": Decimal("0"),
            "commodity_rate": {Decimal("0.09742")},
            "commodity_gbp": Decimal("38.82"),
            "commodity_kwh": Decimal("968.823"),
        },
        "kwh": Decimal("968.823"),
        "net": Decimal("0.00"),
        "vat": Decimal("0.00"),
        "gross": Decimal("0.00"),
        "start_date": utc_datetime(2024, 3, 31, 23, 0),
        "finish_date": utc_datetime(2024, 4, 30, 22, 30),
        "mprn": "978672",
    }

    assert headers == expected_headers



import collections
import datetime
import importlib
import threading
import traceback
from pkgutil import iter_modules

import pytz

from werkzeug.exceptions import BadRequest

import chellow.gas
from chellow.models import (
    BillType,
    GBatch,
    GReadType,
    GSupply,
    GUnit,
    Session,
)

importer_id = 0
import_lock = threading.Lock()
importers = {}

PARSER_PREFIX = "bill_parser_"


def find_parser_names():
    return [
        name[len(PARSER_PREFIX) :]
        for module_finder, name, ispkg in iter_modules(chellow.gas.__path__)
        if name.startswith(PARSER_PREFIX)
    ]


class GBillImporter(threading.Thread):
    def __init__(self, sess, g_batch_id, file_name, file_bytes, parser_name):
        threading.Thread.__init__(self)
        global importer_id
        self.importer_id = importer_id
        importer_id += 1

        self.g_batch_id = g_batch_id
        if len(file_bytes) == 0:
            raise BadRequest("File has zero length")

        self.parser_name = f"bill_parser_{parser_name}"
        try:
            imp_mod = importlib.import_module(f"chellow.gas.{self.parser_name}")
        except ImportError:
            raise BadRequest(
                f"Can't find a parser called '{parser_name}'. Valid parser names are "
                f"{find_parser_names()}."
            )

        self.parser = imp_mod.Parser(file_bytes)
        self.successful_bills = []
        self.failed_bills = []
        self.log = collections.deque()
        self.bill_num = None

    def _log(self, msg):
        with import_lock:
            self.log.appendleft(
                datetime.datetime.now(pytz.utc).strftime("%Y-%m-%d %H:%M:%S")
                + " - "
                + msg
            )

    def status(self):
        return (
            f"Parsed file up to line number: {self.parser.line_number}. Inserting raw "
            f"bills up to bill: {self.bill_num}."
        )

    def run(self):
        try:
            self._log(f"Starting to parse the file with '{self.parser_name}'.")
            with Session() as sess:
                g_batch = GBatch.get_by_id(sess, self.g_batch_id)
                raw_bills = self.parser.make_raw_bills()
                self._log(
                    "Successfully parsed the file, and now I'm starting to insert the "
                    "raw bills."
                )
                for self.bill_num, raw_bill in enumerate(raw_bills):
                    try:
                        bill_type = BillType.get_by_code(
                            sess, raw_bill["bill_type_code"]
                        )
                        g_supply = GSupply.get_by_mprn(sess, raw_bill["mprn"])
                        g_bill = g_batch.insert_g_bill(
                            sess,
                            g_supply,
                            bill_type,
                            raw_bill["reference"],
                            raw_bill["account"],
                            raw_bill["issue_date"],
                            raw_bill["start_date"],
                            raw_bill["finish_date"],
                            raw_bill["kwh"],
                            raw_bill["net_gbp"],
                            raw_bill["vat_gbp"],
                            raw_bill["gross_gbp"],
                            raw_bill["raw_lines"],
                            raw_bill["breakdown"],
                        )
                        sess.flush()
                        for raw_read in raw_bill["reads"]:
                            prev_type = GReadType.get_by_code(
                                sess, raw_read["prev_type_code"]
                            )
                            pres_type = GReadType.get_by_code(
                                sess, raw_read["pres_type_code"]
                            )
                            g_unit = GUnit.get_by_code(sess, raw_read["unit"])
                            g_read = g_bill.insert_g_read(
                                sess,
                                raw_read["msn"],
                                g_unit,
                                raw_read["correction_factor"],
                                raw_read["calorific_value"],
                                raw_read["prev_value"],
                                raw_read["prev_date"],
                                prev_type,
                                raw_read["pres_value"],
                                raw_read["pres_date"],
                                pres_type,
                            )

                            sess.expunge(g_read)
                        self.successful_bills.append(raw_bill)
                        sess.expunge(g_bill)
                    except BadRequest as e:
                        sess.rollback()
                        raw_bill["error"] = e.description
                        self.failed_bills.append(raw_bill)

                if len(self.failed_bills) == 0:
                    sess.commit()
                    self._log(
                        "All the bills have been successfully loaded and attached to "
                        "the batch."
                    )
                else:
                    sess.rollback()
                    self._log(
                        f"The import has finished, but {len(self.failed_bills)} bills "
                        f"failed to load, and so the whole import has been rolled back."
                    )

        except BadRequest as e:
            self._log(e.description)
        except BaseException:
            self._log(f"I've encountered a problem: {traceback.format_exc()}")

    def make_fields(self):
        with import_lock:
            fields = {
                "log": tuple(self.log),
                "is_alive": self.is_alive(),
                "importer_id": self.importer_id,
            }
            if not self.is_alive():
                fields["successful_bills"] = self.successful_bills
                fields["failed_bills"] = self.failed_bills
            return fields


def start_bill_importer(sess, batch_id, file_name, file_bytes, parser_name):
    with import_lock:
        bi = GBillImporter(sess, batch_id, file_name, file_bytes, parser_name)
        importers[bi.importer_id] = bi
        bi.start()

    return bi.importer_id


def get_bill_importer_ids(g_batch_id):
    with import_lock:
        return [k for k, v in importers.items() if v.g_batch_id == g_batch_id]


def get_bill_importer(id):
    with import_lock:
        return importers[id]



import atexit
import collections
import threading
import traceback
from datetime import timedelta

import requests

from sqlalchemy import select

from werkzeug.exceptions import BadRequest

from chellow.models import Contract, RateScript, Session
from chellow.utils import (
    ct_datetime,
    date_format,
    to_utc,
    utc_datetime,
    utc_datetime_now,
    utc_datetime_parse,
)

GLOBAL_ALERT = """There's a problem with the <a
href="/bank_holidays">Bank Holiday importer</a>."""
LAST_RUN_KEY = "last_run"
BH_STATE_KEY = "bh"
DELAY_DAYS = 1

importer = None


def run_import(sess, log_f, set_progress):
    log_f("Starting to check bank holidays")
    contract_name = "bank_holidays"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1996, 4, 1)),
            None,
            {},
        )
        sess.commit()
    contract_props = contract.make_properties()

    if contract_props.get("enabled", True):
        url_str = "https://www.gov.uk/bank-holidays/england-and-wales.ics"

        log_f(f"Downloading from {url_str}.")
        res = requests.get(url_str)
        log_f(" ".join(("Received", str(res.status_code), res.reason)))
        PREFIX = "DTSTART;VALUE=DATE:"
        hols = collections.defaultdict(list)
        for line in res.text.splitlines():
            if line.startswith(PREFIX):
                dt = utc_datetime_parse(line[-8:], "%Y%m%d")
                hols[dt.year].append(dt)

        for year in sorted(hols.keys()):
            year_start = utc_datetime(year, 1, 1)
            rs = sess.scalars(
                select(RateScript).where(
                    RateScript.contract == contract,
                    RateScript.start_date == year_start,
                )
            ).one_or_none()
            if rs is None:
                log_f(
                    f"Adding a new rate script starting at {date_format(year_start)}."
                )

                rs = contract.insert_rate_script(sess, year_start, {})

            script = {"bank_holidays": [v.strftime("%Y-%m-%d") for v in hols[year]]}

            contract.update_rate_script(sess, rs, rs.start_date, rs.finish_date, script)
            sess.commit()
            log_f(f"Updated rate script starting at {date_format(year_start)}.")
    else:
        log_f(
            "The automatic importer is disabled. To enable it, edit the contract "
            "properties to set 'enabled' to True."
        )


class BankHolidayImporter(threading.Thread):
    def __init__(self):
        super(BankHolidayImporter, self).__init__(name="Bank Holiday Importer")
        self.lock = threading.RLock()
        self.messages = collections.deque(maxlen=100)
        self.progress = ""
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def is_locked(self):
        if self.lock.acquire(False):
            self.lock.release()
            return False
        else:
            return True

    def log(self, message):
        self.messages.appendleft(
            f"{utc_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def set_progress(self, progress):
        self.progress = progress

    def run(self):
        while not self.stopped.is_set():
            with Session() as sess:
                try:
                    config = Contract.get_non_core_by_name(sess, "configuration")
                    state = config.make_state()
                    bh_state = state.get(BH_STATE_KEY, {})
                except BaseException as e:
                    msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                    self.log(f"{msg}{traceback.format_exc()}")
                    self.global_alert = GLOBAL_ALERT
                    sess.rollback()

            last_run = bh_state.get(LAST_RUN_KEY)
            if last_run is None or utc_datetime_now() - last_run > timedelta(
                days=DELAY_DAYS
            ):
                self.going.set()

            if self.going.is_set():
                self.global_alert = None
                with Session() as sess:
                    try:
                        config = Contract.get_non_core_by_name(sess, "configuration")
                        state = config.make_state()
                        try:
                            bh_state = state[BH_STATE_KEY]
                        except KeyError:
                            bh_state = state[BH_STATE_KEY] = {}

                        bh_state[LAST_RUN_KEY] = utc_datetime_now()
                        config.update_state(state)
                        sess.commit()
                        run_import(sess, self.log, self.set_progress)
                    except BaseException as e:
                        msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                        self.log(f"{msg}{traceback.format_exc()}")
                        self.global_alert = GLOBAL_ALERT
                        sess.rollback()
                    finally:
                        self.going.clear()
                        self.log("Finished checking bank holidays.")

            else:
                self.log(
                    f"The importer was last run at {date_format(last_run)}. There will "
                    f"be another import when {DELAY_DAYS} days have elapsed since the "
                    f"last run."
                )
                self.going.wait(60 * 60)


def get_importer():
    return importer


def startup():
    global importer
    importer = BankHolidayImporter()
    importer.start()


@atexit.register
def shutdown():
    if importer is not None:
        importer.stop()



from decimal import Decimal

from sqlalchemy import select

from chellow.e.neso import csv_get, parse_date
from chellow.models import Contract, RateScript
from chellow.utils import ct_datetime, to_utc


def hh(supply_source):
    try:
        aahedc_cache = supply_source.caches["aahedc"]
    except KeyError:
        aahedc_cache = supply_source.caches["aahedc"] = {}

    for hh in supply_source.hh_data:
        hh["aahedc-kwh"] = hh["gsp-kwh"]
        try:
            rate = aahedc_cache[hh["start-date"]]
        except KeyError:
            hh_start = hh["start-date"]
            rates = supply_source.non_core_rate("aahedc", hh_start)
            rate = aahedc_cache[hh_start] = float(rates["aahedc_gbp_per_gsp_kwh"])

        hh["aahedc-rate"] = rate
        hh["aahedc-gbp"] = hh["gsp-kwh"] * rate


def neso_import(sess, log, set_progress):
    log("Starting to check for new AAHEDC rates")

    contract_name = "aahedc"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )

    for record in csv_get("ffd29cc8-3c55-4e83-aa0e-73212d4fedba"):
        # {
        #   "_full_text": "'-07':4 '-15':5 '0.012247':7 '0.028 '2026':2 'final':1",
        #   "Published Date": "2025-07-15",
        #   "AAHEDC tariff excluding the Shetland Assistance Amount in p/kwh":
        #   "0.028737",
        #   "Shetland Tariff in p/kwh": "0.012247",
        #   "Year FY": 2026,
        #   "Publication Type": "Final",
        #   "_id": 7,
        #   "Total Scheme Tariff in p/kwh": "0.040984"
        # }

        fy_year = int(record["Year FY"]) - 1
        fy_start = to_utc(ct_datetime(fy_year, 4, 1))
        if fy_start < contract.start_rate_script.start_date:
            continue

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == fy_start,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, fy_start, {})

        rs_script = rs.make_script()
        try:
            rs_record = rs_script["record"]
        except KeyError:
            rs_record = rs_script["record"] = {}

        record_published_date = parse_date(record["Published Date"])

        if (
            "Published Date" not in rs_record
            or parse_date(rs_record["Published Date"]) < record_published_date
        ):
            rs_script["record"] = record
            rs_script["aahedc_gbp_per_gsp_kwh"] = Decimal(
                record["Total Scheme Tariff in p/kwh"]
            ) / Decimal("100")
            rs.update(rs_script)
            sess.commit()

    log("Finished Checking for AAHEDC Tariffs")
    sess.commit()



import csv
from collections import deque
from io import BytesIO, TextIOWrapper

from flask import g, render_template

from sqlalchemy import select

from chellow.e.bill_importer import BillImport, find_parser_names
from chellow.models import (
    Bill,
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_bill_types,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_status(sess, mocker):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    participant = Participant.insert(sess, "hhak", "AK Industries")
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = imp_supplier_contract.insert_batch(sess, "b1", "batch 1", vf)
    bill_import = BillImport(batch)
    bill_import.isAlive = mocker.Mock(return_value=True)
    bill_import.status()


def test_BillImport_run(sess):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "22488", "Water Works")
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    participant = Participant.insert(sess, "hhak", "AK Industries")
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop Ltd", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        "",
        {},
        utc_datetime(2000, 1, 1),
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        utc_datetime(1996, 1, 1),
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    batch = imp_supplier_contract.insert_batch(sess, "b1", "batch 1", vf)
    insert_bill_types(sess)

    data = [
        "N",
        "sla;dkjhgf8u",
        "22 7867 6232 781",
        "sdalhgfuwr",
        "2025-01-01 00:00",
        "2025-03-01 00:00",
        "2025-03-31 23:30",
        "34",
        "81.53",
        "12.89",
        "16.88",
        "{}",
        "Element",
        "nrg",
        "2025-03-01 00:00",
        "2025-03-31 23:30",
        "81.53",
        '{"nrg-kwh": 981}',
    ]

    with BytesIO() as f:
        text_stream = TextIOWrapper(f, write_through=True)
        writer = csv.writer(text_stream)
        writer.writerow(data)
        f.seek(0)

        batch.insert_file(sess, "file 1", f.read(), "csv")
    sess.commit()
    importer = BillImport(batch, now=to_utc(ct_datetime(2025, 1, 1)))
    importer.run()
    sess.rollback()
    assert importer.failed_bills == []
    assert importer.log == deque(
        [
            "2025-01-01 00:00:00 - All the bills have been successfully loaded and "
            "attached to the batch.",
            "2025-01-01 00:00:00 - Starting to parse the file file 1 with 'csv'.",
            "2025-01-01 00:00:00 - Importing bills from batch b1.",
        ]
    )
    bill = sess.scalars(select(Bill)).one()
    assert len(bill.elements) == 1


def test_supplier_bill_import_html(mocker, app):
    with app.app_context():
        with app.test_request_context():
            g.user = None
            g.config = {}
            batch = mocker.Mock()
            failed_bills = [{"error": "MPAN not found"}]
            failed_max_registers = 0
            render_template(
                "/e/supplier_bill_import.html",
                batch=batch,
                failed_bills=failed_bills,
                failed_max_registers=failed_max_registers,
            )


def test_find_parser_names():
    assert len(find_parser_names()) > 0



import csv
from decimal import Decimal
from io import StringIO

from utils import match, match_tables

import chellow.reports.report_429
from chellow.models import (
    BillType,
    Contract,
    GContract,
    GDn,
    GReadingFrequency,
    GUnit,
    MarketRole,
    Participant,
    ReportRun,
    Site,
    User,
    insert_bill_types,
    insert_g_reading_frequencies,
    insert_g_units,
)
from chellow.utils import utc_datetime


def test_process_g_bill_ids(mocker):
    sess = mocker.Mock()
    forecast_date = utc_datetime(2010, 4, 1)

    query = mocker.Mock()
    sess.query.return_value = query
    m_filter = mocker.Mock()
    query.filter.return_value = m_filter
    g_bill = mocker.Mock()
    m_filter.one.return_value = g_bill
    g_bill.g_reads = []

    m_filter.order_by.return_value = []

    g_bill.g_supply = mocker.Mock()
    g_bill.start_date = forecast_date
    g_bill.finish_date = forecast_date

    MockGBill = mocker.patch("chellow.reports.report_429.GBill", autospec=True)
    MockGBill.g_supply = mocker.Mock()
    MockGBill.start_date = forecast_date
    MockGBill.finish_date = forecast_date

    find_g_era_at = g_bill.g_supply.find_g_era_at

    report_context = {}
    g_bill_ids = [1]
    bill_titles = []
    vbf = mocker.Mock()
    titles = []
    csv_writer = mocker.Mock()
    g_contract = mocker.Mock()
    report_run_id = 1

    chellow.reports.report_429._process_g_bill_ids(
        sess,
        report_context,
        g_contract,
        g_bill_ids,
        forecast_date,
        bill_titles,
        vbf,
        titles,
        csv_writer,
        report_run_id,
    )

    find_g_era_at.assert_not_called()


def test_batch(mocker, sess, client):
    vf = utc_datetime(2000, 1, 1)
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_cv_rate_script = {
        "cvs": {
            "EA": {
                1: {"applicable_at": utc_datetime(2020, 10, 3), "cv": 39.2000},
            }
        }
    }
    GContract.insert_industry(sess, "cv", "", {}, vf, None, g_cv_rate_script)
    ug_rate_script = {
        "ug_gbp_per_kwh": {"EA1": Decimal("40.1")},
    }
    GContract.insert_industry(sess, "ug", "", {}, vf, None, ug_rate_script)
    ccl_rate_script = {"ccl_gbp_per_kwh": Decimal("0.00339")}
    GContract.insert_industry(sess, "ccl", "", {}, vf, None, ccl_rate_script)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    charge_script = """
import chellow.gas.ccl
from chellow.gas.engine import g_rates
from chellow.utils import reduce_bill_hhs


def virtual_bill_titles():
    return [
        'units_consumed', 'correction_factor', 'unit_code', 'unit_factor',
        'calorific_value', 'kwh', 'gas_rate', 'gas_gbp', 'ccl_rate',
        'standing_rate', 'standing_gbp', 'net_gbp', 'vat_gbp', 'gross_gbp',
        'problem']


def virtual_bill(ds):
    chellow.gas.ccl.vb(ds)
    for hh in ds.hh_data:
        start_date = hh['start_date']
        bill_hh = ds.bill_hhs[start_date]
        bill_hh['units_consumed'] = hh['units_consumed']
        bill_hh['correction_factor'] = {hh['correction_factor']}
        bill_hh['unit_code'] = {hh['unit_code']}
        bill_hh['unit_factor'] = {hh['unit_factor']}
        bill_hh['calorific_value'] = {hh['calorific_value']}
        kwh = hh['kwh']
        bill_hh['kwh'] = kwh
        gas_rate = float(
            g_rates(ds.sess, ds.caches, db_id, start_date, False)['gas_rate'])
        bill_hh['gas_rate'] = {gas_rate}
        bill_hh['gas_gbp'] = gas_rate * kwh
        bill_hh['ccl_kwh'] = kwh
        ccl_rate = hh['ccl']
        bill_hh['ccl_rate'] = {ccl_rate}
        bill_hh['ccl_kwh'] = kwh
        bill_hh['ccl_gbp'] = kwh * ccl_rate
        if hh['utc_is_month_end']:
            standing_rate = float(
                g_rates(
                    ds.sess, ds.caches, db_id, start_date, False)['standing_rate'])
            bill_hh['standing_rate'] = {standing_rate}
            bill_hh['standing_gbp'] = standing_rate
        if hh['utc_decimal_hour'] == 0:
            pass

        bill_hh['net_gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat_gbp'] = 0
        bill_hh['gross_gbp'] = bill_hh['net_gbp'] + bill_hh['vat_gbp']

    ds.bill = reduce_bill_hhs(ds.bill_hhs)
"""
    g_contract_rate_script = {
        "gas_rate": 0.1,
        "standing_rate": 0.1,
    }
    g_contract = GContract.insert(
        sess,
        False,
        "Fusion 2020",
        charge_script,
        {},
        utc_datetime(2000, 1, 1),
        None,
        g_contract_rate_script,
    )
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch", vf)
    g_batch_id = g_batch.id

    breakdown = {"units_consumed": 771}
    insert_bill_types(sess)
    bill_type_n = BillType.get_by_code(sess, "N")
    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )
    g_bill_id = g_bill.id
    user = User.get_by_email_address(sess, "admin@example.com")
    user_id = user.id
    report_run = ReportRun.insert(sess, "g_bill_check", None, "", {})
    report_run_id = report_run.id
    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_429.open_file", return_value=mock_file)

    chellow.reports.report_429.content(
        g_batch_id, g_bill_id, None, None, None, user_id, report_run_id
    )

    mock_file.seek(0)
    actual = list(csv.reader(mock_file))
    expected = [
        [
            "batch",
            "bill_reference",
            "bill_type",
            "bill_start_date",
            "bill_finish_date",
            "mprn",
            "supply_name",
            "site_code",
            "site_name",
            "covered_start",
            "covered_finish",
            "covered_bill_ids",
            "covered_units_consumed",
            "virtual_units_consumed",
            "covered_correction_factor",
            "virtual_correction_factor",
            "" "covered_unit_code",
            "virtual_unit_code",
            "covered_unit_factor",
            "virtual_unit_factor",
            "covered_calorific_value",
            "virtual_calorific_value",
            "covered_kwh",
            "virtual_kwh",
            "covered_gas_rate",
            "virtual_gas_rate",
            "covered_gas_gbp",
            "virtual_gas_gbp",
            "difference_gas_gbp",
            "covered_ccl_rate",
            "virtual_ccl_rate",
            "covered_standing_rate",
            "virtual_standing_rate",
            "covered_standing_gbp",
            "virtual_standing_gbp",
            "difference_standing_gbp",
            "covered_net_gbp",
            "virtual_net_gbp",
            "difference_net_gbp",
            "covered_vat_gbp",
            "virtual_vat_gbp",
            "difference_vat_gbp",
            "covered_gross_gbp",
            "virtual_gross_gbp",
            "difference_gross_gbp",
            "covered_problem",
            "virtual_problem",
        ],
        [
            "b1",
            "55h883",
            "N",
            "2020-01-01 00:00",
            "2020-01-31 23:30",
            "87614362",
            "main",
            "22488",
            "Water Works",
            "2020-01-01 00:00",
            "2020-01-31 23:30",
            "1",
            "771",
            "0",
            "",
            "1.0",
            "",
            "M3",
            "",
            "1.0",
            "",
            "39.2",
            "45",
            "0.0",
            "",
            "0.1",
            "",
            "0.0",
            "0.0",
            "",
            "0.00339",
            "",
            "0.1",
            "",
            "0.1",
            "-0.1",
            "12.40",
            "0.1",
            "12.3",
            "1.20",
            "0",
            "" "1.2",
            "14.52",
            "0.1",
            "14.42",
            "",
            "",
        ],
    ]
    match_tables(expected, actual)


def test_batch_http(mocker, sess, client):
    vf = utc_datetime(2000, 1, 1)
    g_contract = GContract.insert(sess, False, "Fusion 2020", "", {}, vf, None, {})
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch", vf)

    user = User.get_by_email_address(sess, "admin@example.com")
    sess.commit()

    query_string = {"g_batch_id": g_batch.id}

    mock_Thread = mocker.patch(
        "chellow.reports.report_429.threading.Thread", autospec=True
    )
    response = client.get("/reports/429", query_string=query_string)

    match(response, 303)

    report_run_id = 1

    args = g_batch.id, None, None, None, None, user.id, report_run_id
    mock_Thread.assert_called_with(target=chellow.reports.report_429.content, args=args)


def test_bill_http(mocker, sess, client):
    vf = utc_datetime(2000, 1, 1)
    site = Site.insert(sess, "22488", "Water Works")
    g_dn = GDn.insert(sess, "EE", "East of England")
    g_ldz = g_dn.insert_g_ldz(sess, "EA")
    g_exit_zone = g_ldz.insert_g_exit_zone(sess, "EA1")
    insert_g_units(sess)
    g_unit_M3 = GUnit.get_by_code(sess, "M3")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    g_contract = GContract.insert_supplier(sess, "Fusion 2020", "", {}, vf, None, {})
    insert_g_reading_frequencies(sess)
    g_reading_frequency_M = GReadingFrequency.get_by_code(sess, "M")
    g_supply = site.insert_g_supply(
        sess,
        "87614362",
        "main",
        g_exit_zone,
        utc_datetime(2018, 1, 1),
        None,
        "hgeu8rhg",
        1,
        g_unit_M3,
        g_contract,
        "d7gthekrg",
        g_reading_frequency_M,
        1,
        1,
    )
    g_batch = g_contract.insert_g_batch(sess, "b1", "Jan batch", vf)

    breakdown = {"units_consumed": 771}
    insert_bill_types(sess)
    bill_type_n = BillType.get_by_code(sess, "N")
    g_bill = g_batch.insert_g_bill(
        sess,
        g_supply,
        bill_type_n,
        "55h883",
        "dhgh883",
        utc_datetime(2019, 4, 3),
        utc_datetime(2020, 1, 1),
        utc_datetime(2020, 1, 31, 23, 30),
        Decimal("45"),
        Decimal("12.40"),
        Decimal("1.20"),
        Decimal("14.52"),
        "",
        breakdown,
    )
    user = User.get_by_email_address(sess, "admin@example.com")
    sess.commit()

    query_string = {"g_bill_id": g_bill.id}

    mock_Thread = mocker.patch(
        "chellow.reports.report_429.threading.Thread", autospec=True
    )
    response = client.get("/reports/429", query_string=query_string)

    match(response, 303)

    report_run_id = 1

    args = None, g_bill.id, None, None, None, user.id, report_run_id
    mock_Thread.assert_called_with(target=chellow.reports.report_429.content, args=args)



import csv
import threading
import traceback
from datetime import datetime as Datetime, timedelta as Timedelta

from flask import g, redirect

import pytz

from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from chellow.dloads import open_file
from chellow.models import Session, Site, Snag, User
from chellow.utils import date_format


def content(user_id):
    f = writer = None
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("site_snags.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            writer.writerow(
                (
                    "Chellow Id",
                    "Site Code",
                    "Site Name",
                    "Snag Description",
                    "Start Date",
                    "Finish Date",
                    "Days Since Snag Finished",
                    "Duration Of Snag (Days)",
                    "Is Ignored?",
                )
            )

            now = Datetime.now(pytz.utc)

            for snag in (
                sess.query(Snag)
                .join(Site)
                .filter(Snag.site != null())
                .order_by(Site.code, Snag.description, Snag.start_date, Snag.id)
                .options(joinedload(Snag.site))
            ):
                snag_start = snag.start_date
                snag_finish = snag.finish_date
                if snag_finish is None:
                    duration = now - snag_start
                    age_of_snag = Timedelta(0)
                else:
                    duration = snag_finish - snag_start
                    age_of_snag = now - snag_finish

                writer.writerow(
                    (
                        str(snag.id),
                        snag.site.code,
                        snag.site.name,
                        snag.description,
                        date_format(snag_start),
                        date_format(snag_finish),
                        str(age_of_snag.days + age_of_snag.seconds / (3600 * 24)),
                        str(duration.days + duration.seconds / (3600 * 24)),
                        str(snag.is_ignored),
                    )
                )
    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        if writer is not None:
            writer.writerow([msg])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    args = (g.user.id,)
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



import csv
from decimal import Decimal

from io import BytesIO, TextIOWrapper

from chellow.e.bill_parsers.csv import Parser, _process_row
from chellow.utils import ct_datetime, to_utc


def test_process_row():
    vals = [
        "N",
        "SA342376000",
        "22 1065 3921 534",
        "3423760004",
        "2010-02-02 00:00",
        "2010-01-05 00:00",
        "2010-01-10 23:30",
        "150",
        "98.174",  # Check rounded to two dp
        "15.01",
        "0",
        "",
        "read",
        "I02D89150",
        "22 1065 3921 534",
        "1",
        "kWh",
        "1",
        "2010-01-04 23:30",
        "14281",
        "E",
        "2010-01-06 23:30",
        "15924",
        "N",
        "element",
        "nrg",
        "2010-01-06 00:00",
        "2010-01-11 23:30",
        "5.123",
        "{}",
    ]
    actual = _process_row(vals)
    expected = {
        "account": "SA342376000",
        "bill_type_code": "N",
        "breakdown": {},
        "elements": [
            {
                "breakdown": {},
                "finish_date": to_utc(ct_datetime(2010, 1, 11, 23, 30)),
                "name": "nrg",
                "net": Decimal("5.12"),
                "start_date": to_utc(ct_datetime(2010, 1, 6, 0, 0)),
            },
        ],
        "finish_date": to_utc(ct_datetime(2010, 1, 10, 23, 30)),
        "gross": Decimal("0.00"),
        "issue_date": to_utc(ct_datetime(2010, 2, 2, 0, 0)),
        "kwh": Decimal("150"),
        "mpan_core": "22 1065 3921 534",
        "net": Decimal("98.17"),
        "reads": [
            {
                "coefficient": Decimal("1"),
                "mpan": "22 1065 3921 534",
                "msn": "I02D89150",
                "pres_date": to_utc(ct_datetime(2010, 1, 6, 23, 30)),
                "pres_type_code": "N",
                "pres_value": Decimal("15924"),
                "prev_date": to_utc(ct_datetime(2010, 1, 4, 23, 30)),
                "prev_type_code": "E",
                "prev_value": Decimal("14281"),
                "tpr_code": "00001",
                "units": "kWh",
            },
        ],
        "reference": "3423760004",
        "start_date": to_utc(ct_datetime(2010, 1, 5, 0, 0)),
        "vat": Decimal("15.01"),
    }
    assert actual == expected


def test_bill_parser_csv():
    """
    Check bills have a UTC timezone
    """

    vals = [
        [
            "#InvoiceType",
            "Account Reference",
            "Mpans",
            "Invoice Reference",
            "Issue Date",
            "Start Date",
            "Finish Date",
            "kWh",
            "Net",
            "VAT",
            "Gross",
            "Breakdown",
            "Record Type",
            "R1 Meter Serial Number",
            "R1 MPAN",
            "R1 Coefficient",
            "R1 Units",
            "R1 TPR",
            "R1 Previous Read Date",
            "R1 Previous Read Value",
            "R1 Previous Read Type",
            "R1 Present Read Date",
            "R1 Present Read Value",
            "R1 Present Read Type",
        ],
        [
            "N",
            "SA342376000",
            "22 1065 3921 534",
            "3423760004",
            "2010-02-02 00:00",
            "2010-01-05 00:00",
            "2010-01-10 23:30",
            "150",
            "98.17",
            "15.01",
            "0",
            "",
            "read",
            "I02D89150",
            "22 1065 3921 534",
            "1",
            "kWh",
            "1",
            "2010-01-04 23:30",
            "14281",
            "E",
            "2010-01-06 23:30",
            "15924",
            "N",
        ],
        [
            "N",
            "SA342376000",
            "22 1065 3921 534",
            "3423760005",
            "2011-02-02 00:00",
            "2011-01-05 00:00",
            "2011-01-10 23:30",
            "150",
            "98.17",
            "15.01",
            "0",
            "",
            "read",
            "I02D89150",
            "22 1065 3921 534",
            "1",
            "kWh",
            "1",
            "2011-01-04 23:30",
            "24286",
            "E",
            "2011-01-06 23:30",
            "25927",
            "E",
            "read",
            "I02D89150",
            "22 1065 3921 534",
            "1",
            "kWh",
            "1",
            "2011-02-04 23:30",
            "34285",
            "E",
            "2011-02-06 23:30",
            "46883",
            "E",
            "read",
            "I02D89150",
            "22 1065 3921 534",
            "1",
            "kWh",
            "3",
            "2011-02-04 23:30",
            "8553",
            "E",
            "2011-02-06 23:30",
            "8553",
            "E",
        ],
    ]
    with BytesIO() as f:
        writer = csv.writer(TextIOWrapper(f, write_through=True))
        writer.writerows(vals)
        f.seek(0)
        parser = Parser(f)
        for bill in parser.make_raw_bills():
            for read in bill["reads"]:
                for k in ("prev_date", "pres_date"):
                    assert read[k].tzinfo is not None


def test_make_raw_bills_whitespace_comment():
    """
    Check bills have a UTC timezone
    """

    vals = [
        [
            "\ufeff#InvoiceType",  # Check whitespace is ignored
            "Account Reference",
            "Mpans",
            "Invoice Reference",
            "Issue Date",
            "Start Date",
            "Finish Date",
            "kWh",
            "Net",
            "VAT",
            "Gross",
            "Breakdown",
            "Record Type",
            "R1 Meter Serial Number",
            "R1 MPAN",
            "R1 Coefficient",
            "R1 Units",
            "R1 TPR",
            "R1 Previous Read Date",
            "R1 Previous Read Value",
            "R1 Previous Read Type",
            "R1 Present Read Date",
            "R1 Present Read Value",
            "R1 Present Read Type",
        ],
    ]
    with BytesIO() as f:
        writer = csv.writer(TextIOWrapper(f, write_through=True))
        writer.writerows(vals)
        f.seek(0)
        parser = Parser(f)
        for bill in parser.make_raw_bills():
            pass



from decimal import Decimal
from io import BytesIO, StringIO
from zipfile import BadZipFile, ZipFile

import pytest

from chellow.e.lafs import _process, find_participant_entries, laf_days
from chellow.models import (
    Contract,
    MarketRole,
    Participant,
    VoltageLevel,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_laf_days(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    participant_code = "IPNL"
    participant = Participant.insert(sess, participant_code, "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    calb_participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    calb_participant.insert_party(sess, market_role_Z, "NonCore", vf, None, "")
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(sess, "500", "PC 5-8 & HH HV", voltage_level, False, True, vf, None)
    sess.commit()
    csv_file = StringIO("""ZHD||D0265001|R|IPNL|G|CAPG|20210914142620||||OPER
DIS|IPNL
LLF|500
SDT|20210922
SPL|1|1.074
ZPT|40672767|423780412""")
    csv_dt = utc_datetime(2021, 9, 1)
    log = mocker.Mock()
    set_progress = mocker.Mock()
    actual = list(laf_days(sess, log, set_progress, csv_file, csv_dt))
    expected = [([1], [utc_datetime(2021, 9, 21, 23, 0)], [Decimal("1.074")])]
    assert actual == expected


def test_process_empty_file(mocker, sess):
    log = mocker.Mock()
    set_progress = mocker.Mock()
    f = BytesIO()
    zf = ZipFile(f, mode="w")
    zf.writestr("llfipnl20210922.ptf", "")
    zf.close()
    f.seek(0)
    _process(sess, log, set_progress, f)


def test_process(mocker, sess):
    vf = to_utc(ct_datetime(2000, 1, 1))
    participant_code = "IPNL"
    participant = Participant.insert(sess, participant_code, "AK Industries")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    calb_participant = Participant.insert(sess, "CALB", "AK Industries")
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    calb_participant.insert_party(sess, market_role_Z, "NonCore", vf, None, "")
    Contract.insert_non_core(sess, "configuration", "", {}, vf, None, {})
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    dno.insert_llfc(sess, "500", "PC 5-8 & HH HV", voltage_level, False, True, vf, None)
    sess.commit()
    log = mocker.Mock()
    set_progress = mocker.Mock()
    csv_str = """ZHD||D0265001|R|IPNL|G|CAPG|20210914142620||||OPER
DIS|IPNL
LLF|500
SDT|20210922
SPL|1|1.074
ZPT|40672767|423780412"""
    f = BytesIO()
    zf = ZipFile(f, mode="w")
    zf.writestr("llfipnl20210922.ptf", csv_str)
    zf.close()
    f.seek(0)
    _process(sess, log, set_progress, f)


def test_laf_import_error(mocker, sess):
    file_lines = ("",)

    file_bytes = "\n".join(file_lines).encode("utf8")
    f = BytesIO(file_bytes)

    log = mocker.Mock()
    set_progress = mocker.Mock()

    with pytest.raises(BadZipFile, match="File is not a zip file"):
        _process(sess, log, set_progress, f)


def test_find_participant_entries_empty():
    paths = []
    laf_cache = {}
    actual_entries = find_participant_entries(paths, laf_cache)
    expected_entries = {}
    assert actual_entries == expected_entries


def test_find_participant_entries_single():
    paths = [(("2023", "electricity", "lafs", "llfipnl20210922.ptf.zip"), "url")]
    laf_cache = {}
    actual_entries = find_participant_entries(paths, laf_cache)
    expected_entries = {"ipnl": {"20210922": "url"}}
    assert actual_entries == expected_entries


def test_find_participant_entries_single_in_state():
    paths = [(("2023", "electricity", "lafs", "llfipnl20210922.ptf.zip"), "url")]
    laf_cache = {"ipnl": "20210922"}
    actual_entries = find_participant_entries(paths, laf_cache)
    expected_entries = {}
    assert actual_entries == expected_entries
